/**
 * @file   mycp.c
 * @author Jade Cheng
 * @date   September 14, 2009
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

/*---------------------------------------------------------------------------*/
/**
 * Copy a file using a specified buffer size.
 *
 * @param bufsize The specified buffer size.
 * @param source The source file to be copied.
 * @param destination The desitnation file to be formed.
 *
 * @return EXIT_SUCCESS or EXIT_FAILURE.
 */
static int copy_file(
	size_t       bufsize,
	const char * source,
	const char * destination);

/**
 * The main entry point of the program.
 *
 * @param argc The number of arguments.
 * @param argv The command line arguments.
 *
 * @return EXIT_SUCCESS or EXIT_FAILURE.
 */
extern int main(int argc, const char ** argv)
{
	size_t bufsize;
	
	/* Check for --help argument and display usage. */
	if (argc == 2 && 0 == strcmp(argv[1], "--help")) {
		printf("Usage: mycp buffersize source destination\n");
		return EXIT_SUCCESS;
	}

	/* Check if the right number of arguments were entered. */
	if (argc != 4) {
		fprintf(stderr, "Arguments are invalid.  "
			"Enter \"./mycp --help\" for usage.\n");
		return EXIT_FAILURE;
	}

	/* Check if the buffer size is valid, and obtain this number. */
	if (1 != sscanf(argv[1], "%u", &bufsize) ||
			argv[1][0] == '-' ||
			bufsize == 0) {
		fprintf(stderr, "Buffersize: %s is invalid\n", argv[1]);
		return EXIT_FAILURE;
	}

	/* Call copy_file to copyi with the parsed arguments. */
	return copy_file(bufsize, argv[2], argv[3]);
}

/*---------------------------------------------------------------------------*/
static int copy_file(size_t bufsize, const char * source, const char * destination)
{
	char * buf;
	int    dst;
	int    src;

	/* Assert the parameters are valid. */
	assert(bufsize > 0);
	assert(NULL != source);
	assert(NULL != destination);

	/* Allocate memory with the specified size. */
	if (NULL == (buf = malloc(bufsize))) {
		fprintf(stderr, "Failed to malloc %u bytes of memory\n",
			bufsize);
		return EXIT_FAILURE;
	}

	/* Open source file to read. */
	if (0 >= (src = open(source, O_RDONLY))) {
		fprintf(stderr, "Failed to open file: %s\n", source);
	        free(buf);
		return EXIT_FAILURE;
	}

	/* Create destination file to write or rewrite. */
	if (0 >= (dst = open(destination, O_WRONLY | O_CREAT | O_TRUNC))) {
		fprintf(stderr, "Failed to create file: %s\n", destination);
		close(src);
		free(buf);
		return EXIT_FAILURE;
	}

	/* Copy a chunk at a time until the entire file is copied. */
	for(;;) {
		size_t actual;

		/* Exit the loop if the end of the source file is reached. */
		if (0 == (actual = read(src, buf, sizeof(char) * bufsize)))
			break;

		/* Return a failure if an error occurs while reading. */
		if (actual < 0) {
			fprintf(stderr, "Failed to read from file: %s\n.",
				source);
			close(dst);
			close(src);
			free(buf);
			return EXIT_FAILURE;
		}

		/* Return a failure if an error occurs while writing. */
		if (actual != (write(dst, buf, sizeof(char) * actual))) {
			fprintf(stderr, "Failed to write to file: %s.\n",
				destination);
			close(dst);
			close(src);
			free(buf);
			return EXIT_FAILURE;
		}
	}

	/* Close the files and free the buffer memory. */
	close(dst);
	close(src);
	free(buf);
	return EXIT_SUCCESS;
}

